from typing import Dict, Optional, Set, Tuple, Type

import networkx as nx
import pandas as pd
import pytest

from visions import VisionsBaseType, VisionsTypeset


def all_series_included(series_list, series_map):
    """Check that all names are indeed used"""
    used_names = set([name for names in series_map.values() for name in names])
    names = set([series.name for series in series_list])
    if not names == used_names:
        unused = names ^ used_names
        # TODO: warning?
        raise ValueError(f"Not all series are included {unused}")


def get_contains_cases(
    _test_suite, _series_map: Dict[Type[VisionsBaseType], Set[str]], typeset
):
    """Parametrize contains tests

    Args:
        mapping: mapping from type to a set of series' identifiers

    Returns:
        the args for the generated tests
    """

    # Include children's series in parent
    reversed_topological_edge_sort = list(
        reversed(list(nx.topological_sort(nx.line_graph(typeset.base_graph))))
    )
    for parent, child in reversed_topological_edge_sort:
        _series_map[parent] |= _series_map[child]

    all_series_included(_test_suite, _series_map)

    argsvalues = []
    for item in _test_suite:
        for type, series_list in _series_map.items():
            args = {"id": f"{item.name} x {type}"}

            member = item.name in series_list
            argsvalues.append(pytest.param(item, type, member, **args))

    return {"argnames": "series,type,member", "argvalues": argsvalues}


def contains(series, type, member):
    return (
        member == (series in type),
        f"{series.name} in {type}; expected {member}, got {series in type}",
    )


def get_inference_cases(_test_suite, inferred_series_type_map, typeset):
    argsvalues = []
    for series in _test_suite:
        expected_type = inferred_series_type_map[series.name]
        for test_type in typeset.types:
            expected = test_type == expected_type
            args = {"id": f"{series.name} x {test_type} expected {expected}"}
            difference = test_type != expected_type
            argsvalues.append(
                pytest.param(series, test_type, typeset, difference, **args)
            )
    return {"argnames": "series,type,typeset,difference", "argvalues": argsvalues}


def infers(series, expected_type, typeset, difference):
    # TODO: include paths on error!
    inferred_type = typeset.infer_type(series)
    return (
        (inferred_type == expected_type) != difference,
        f"inference of {series.name} expected {expected_type} to be {not difference} (typeset={typeset})",
    )
    # return series in inferred_type, f"series should be member of inferred type"


def all_relations_tested(series_map, typeset):
    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            f"Not all inferential relations are tested {missing_relations}"
        )


def get_convert_cases(_test_suite, _series_map, typeset):
    all_relations_tested(_series_map, typeset)

    argsvalues = []
    for item in _test_suite:
        for source_type, relation_type, series_list in _series_map:
            if item in relation_type:
                args = {"id": f"{item.name}: {relation_type} -> {source_type}"}
                member = item.name in series_list
                argsvalues.append(
                    pytest.param(source_type, relation_type, item, member, **args)
                )

    return dict(
        argnames=["source_type", "relation_type", "series", "member"],
        argvalues=argsvalues,
    )


def convert(source_type, relation_type, series, member) -> Tuple[bool, str]:
    relation_gen = (
        rel for rel in source_type.relations if rel.related_type == relation_type
    )
    try:
        relation = next(relation_gen)
        is_relation = relation.is_relation(series, {})
    except StopIteration:
        relation = None
        is_relation = False

    if not member:
        return (
            (not is_relation),
            f"{source_type}, {relation}, {member}, {series.name}, {series[0]}",
        )
    else:
        # Note that the transformed series is not exactly the cast series
        transformed_series = relation.transform(series, {})

        return (
            is_relation,
            f"Relationship {relation} transformed {series.values} to {transformed_series.values}",
        )


def get_cast_cases(_test_suite, _results):
    argsvalues = []
    for item in _test_suite:
        changed = item.name in _results
        value = _results.get(item.name, "")
        args = {"id": f"{item.name}: {changed}"}
        argsvalues.append(pytest.param(item, value, **args))

    return dict(
        argnames=["series", "expected"],
        argvalues=argsvalues,
    )


def cast(
    series: pd.Series, typeset: VisionsTypeset, expected: Optional[pd.Series] = None
):
    result = typeset.cast_to_inferred(series)
    # TODO: if error also print Path
    if expected is None:
        v = result.equals(series)
        m = f"Series {series.name} cast expected {series.values} (dtype={series.dtype}) (no casting) got {result.values} (dtype={result.dtype})"

        if v:
            v = id(series) == id(result)
            m = f"Series {series.name} memory addresses are not equal, while return value was"
    else:
        v = result.equals(expected)
        m = f"Series {series.name} cast expected {expected.values} (dtype={expected.dtype}) got {result.values} (dtype={result.dtype})"

    return v, m



from collections import Counter
from typing import Dict

import pandas as pd
from tangled_up_in_unicode import block, block_abbr, category, category_long, script

from visions.application.summaries.series.numerical_summary import (
    named_aggregate_summary,
)


def get_character_counts(series: pd.Series) -> Counter:
    """Function to return the character counts

    Args:
        series: the Series to process

    Returns:
        A dict with character counts
    """
    return Counter(series.str.cat())


def counter_to_series(counter: Counter) -> pd.Series:
    if not counter:
        return pd.Series()

    counter_as_tuples = counter.most_common()
    items, counts = zip(*counter_as_tuples)
    return pd.Series(counts, index=items)


def unicode_summary(series: pd.Series) -> dict:
    # Unicode Character Summaries (category and script name)
    character_counts = get_character_counts(series)

    character_counts_series = counter_to_series(character_counts)

    char_to_block = {key: block(key) for key in character_counts.keys()}
    char_to_category_short = {key: category(key) for key in character_counts.keys()}
    char_to_script = {key: script(key) for key in character_counts.keys()}

    summary = {
        "n_characters": len(character_counts_series),
        "character_counts": character_counts_series,
        "category_alias_values": {
            key: category_long(value) for key, value in char_to_category_short.items()
        },
        "block_alias_values": {
            key: block_abbr(value) for key, value in char_to_block.items()
        },
    }

    # Retrieve original distribution
    block_alias_counts: Counter = Counter()
    per_block_char_counts: dict = {
        k: Counter() for k in summary["block_alias_values"].values()
    }
    for char, n_char in character_counts.items():
        block_name = summary["block_alias_values"][char]
        block_alias_counts[block_name] += n_char
        per_block_char_counts[block_name][char] = n_char
    summary["block_alias_counts"] = counter_to_series(block_alias_counts)
    summary["block_alias_char_counts"] = {
        k: counter_to_series(v) for k, v in per_block_char_counts.items()
    }

    script_counts: Counter = Counter()
    per_script_char_counts: dict = {k: Counter() for k in char_to_script.values()}
    for char, n_char in character_counts.items():
        script_name = char_to_script[char]
        script_counts[script_name] += n_char
        per_script_char_counts[script_name][char] = n_char
    summary["script_counts"] = counter_to_series(script_counts)
    summary["script_char_counts"] = {
        k: counter_to_series(v) for k, v in per_script_char_counts.items()
    }

    category_alias_counts: Counter = Counter()
    per_category_alias_char_counts: dict = {
        k: Counter() for k in summary["category_alias_values"].values()
    }
    for char, n_char in character_counts.items():
        category_alias_name = summary["category_alias_values"][char]
        category_alias_counts[category_alias_name] += n_char
        per_category_alias_char_counts[category_alias_name][char] += n_char
    summary["category_alias_counts"] = counter_to_series(category_alias_counts)
    summary["category_alias_char_counts"] = {
        k: counter_to_series(v) for k, v in per_category_alias_char_counts.items()
    }

    # Unique counts
    summary["n_category"] = len(summary["category_alias_counts"])
    summary["n_scripts"] = len(summary["script_counts"])
    summary["n_block_alias"] = len(summary["block_alias_counts"])

    return summary


def length_summary(series: pd.Series) -> dict:
    length = series.str.len()
    summary = {"length": length}
    summary.update(named_aggregate_summary(length, "length"))
    return summary


def text_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    # Distribution of length
    summary = {}
    summary.update(length_summary(series))
    summary.update(unicode_summary(series))

    return summary



import pandas as pd

from examples.data_analysis.categorical import Category
from visions.functional import detect_type
from visions.types import Boolean, Categorical
from visions.typesets import StandardSet

ts = StandardSet()
ts -= Boolean
ts -= Categorical
ts += Category

s1 = pd.Series(["A", "B", "C"] * 1000, dtype="category")
print(s1 in Category)
print(detect_type(s1, ts))

s2 = pd.Series([True, False] * 1000)
print(s2 in Category)
print(detect_type(s2, ts))



import os
import sys
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import nullable_series_contains


def string_is_geometry(series: pd.Series, state: dict) -> bool:
    """Shapely logs failures at a silly severity, just trying to suppress it's output on failures."""
    from shapely import wkt
    from shapely.errors import WKTReadingError

    # only way to get rid of sys output when wkt.loads hits a bad value
    # TODO: use coercion wrapper for this
    sys.stderr = open(os.devnull, "w")
    try:
        result = all(wkt.loads(value) for value in series)
    except (WKTReadingError, AttributeError, UnicodeEncodeError, TypeError):
        result = False
    finally:
        sys.stderr = sys.__stderr__
    return result


def to_geometry(series: pd.Series, state: dict) -> pd.Series:
    from shapely import wkt

    return pd.Series([wkt.loads(value) for value in series])


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(
            cls, String, relationship=string_is_geometry, transformer=to_geometry
        ),
    ]
    return relations


# TODO: Evaluate https://jorisvandenbossche.github.io/blog/2019/08/13/geopandas-extension-array-refactor/
class Geometry(VisionsBaseType):
    """**Geometry** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> from shapely import wkt
        >>> x = pd.Series([wkt.loads('POINT (-92 42)'), wkt.loads('POINT (-92 42.1)'), wkt.loads('POINT (-92 42.2)')]
        >>> x in visions.geometry
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        from shapely.geometry.base import BaseGeometry

        return all(issubclass(type(x), BaseGeometry) for x in series)



from functools import partial
from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils


def string_is_datetime(series: pd.Series, state: dict) -> bool:
    exceptions = [OverflowError, TypeError]
    return test_utils.coercion_test(partial(to_datetime, state=state), exceptions)(
        series
    )


def to_datetime(series: pd.Series, state: dict) -> pd.Series:
    return pd.to_datetime(series)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=string_is_datetime,
            transformer=to_datetime,
        ),
    ]
    return relations


class DateTime(VisionsBaseType):
    """**Datetime** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([pd.datetime(2017, 3, 5), pd.datetime(2019, 12, 4)])
        >>> x in visions.DateTime
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_datetime64_any_dtype(series)



from typing import Dict, List, Tuple, Type, Union

import pandas as pd

from visions.types.type import VisionsBaseType
from visions.typesets.typeset import VisionsTypeset


def cast_to_detected(
    data: Union[pd.Series, pd.DataFrame], typeset: VisionsTypeset
) -> pd.DataFrame:
    """Casts a DataFrame into a typeset by first performing column wise type inference against
    a provided typeset

    Args:
        df: the DataFrame to cast
        typeset: the Typeset in which we cast

    Returns:
        A tuple of the casted DataFrame and the types to which the columns were cast
    """
    return typeset.cast_to_detected(data)


def cast_to_inferred(
    data: Union[pd.Series, pd.DataFrame], typeset: VisionsTypeset
) -> Tuple[pd.DataFrame, dict]:
    """Casts a DataFrame into a typeset by first performing column wise type inference against
    a provided typeset

    Args:
        df: the DataFrame to cast
        typeset: the Typeset in which we cast

    Returns:
        A tuple of the casted DataFrame and the types to which the columns were cast
    """
    return typeset.cast_to_inferred(data)


def infer_type(
    data: Union[pd.Series, pd.DataFrame], typeset: VisionsTypeset
) -> Union[Dict[str, Type[VisionsBaseType]], Type[VisionsBaseType]]:
    """Infer the current types of each column in the DataFrame given the typeset.

    Args:
        df: the DataFrame to infer types on
        typeset: the Typeset that provides the type context

    Returns:
        A dictionary with a mapping from column name to type
    """
    return typeset.infer_type(data)


def detect_type(
    data: Union[pd.Series, pd.DataFrame], typeset: VisionsTypeset
) -> Union[Dict[str, Type[VisionsBaseType]], Type[VisionsBaseType]]:
    """Detect the type in the base graph

    Args:
        df: the DataFrame to detect types on
        typeset: the Typeset that provides the type context

    Returns:
        A dictionary with a mapping from column name to type
    """
    return typeset.detect_type(data)


def compare_detect_inference_frame(
    df: pd.DataFrame, typeset: VisionsTypeset
) -> List[Tuple[str, Type[VisionsBaseType], Type[VisionsBaseType]]]:
    """Compare the types given by inference on the base graph and the relational graph

    Args:
        df: the DataFrame to detect types on
        typeset: the Typeset that provides the type context

    Examples:
        >>> for column, type_before, type_after in compare_detect_inference_frame(df, typeset):
        >>>    print(f"{column} was {type_before} is {type_after}")

    See Also:
        :doc:`type_inference_report_frame <visions.functional.type_inference_report_frame>`: Formatted report of the output of this function
    """
    comparisons = []
    detected_types = detect_type(df, typeset)
    inferred_types = infer_type(df, typeset)
    for key in detected_types.keys() & inferred_types.keys():  # type: ignore
        comparisons.append(
            (key, detected_types[key], inferred_types[key])  # type: ignore
        )
    return comparisons


def type_inference_report_frame(df, typeset) -> str:
    """Return formatted report of the output of `compare_detect_inference_frame`.

    Args:
        df: the DataFrame to detect types on
        typeset: the Typeset that provides the type context

    Returns:
        Text-based comparative type inference report

    Examples:
        >>> import pandas as pd
        >>> from visions.functional import type_inference_report_frame
        >>> from visions.typesets import StandardSet
        >>>
        >>> typeset = StandardSet()
        >>> df = pd.read_csv('dataset.csv')
        >>>
        >>> report = type_inference_report_frame(df, typeset)
        >>> print(report)
    """
    padding = 5
    max_column_length = max([len(column) for column in df.columns]) + padding
    max_type_length = 30

    report = ""
    change_count = 0
    for column, type_before, type_after in compare_detect_inference_frame(df, typeset):
        changed = type_before != type_after
        if changed:
            fill = "!="
            change_count += 1
        else:
            fill = "=="
        report += "{column: <{max_column_length}} {type_before: <{max_type_length}} {fill} {type_after: <{max_type_length}} \n".format(
            column=column,
            max_column_length=max_column_length,
            type_before=str(type_before),
            type_after=str(type_after),
            max_type_length=max_type_length,
            fill=fill,
        )
    report += (
        "In total {change_count} out of {type_count} types were changed.\n".format(
            change_count=change_count, type_count=len(df.columns)
        )
    )
    return report



import pytest

from visions import EmailAddress
from visions.test.series import get_geometry_series, get_series
from visions.types import Generic
from visions.typesets import CompleteSet

typeset = CompleteSet()
typeset += EmailAddress


def pytest_generate_tests(metafunc):
    _test_suite = get_series() + get_geometry_series()
    if metafunc.function.__name__ in ["test_consistency", "test_traversal_mutex"]:
        argsvalues = []
        for series in _test_suite:
            args = {"id": series.name}
            argsvalues.append(pytest.param(series, **args))

        metafunc.parametrize(argnames=["series"], argvalues=argsvalues)


def test_consistency(series):
    detected_type = typeset.detect_type(series)
    message = f"Detected type {detected_type} for series {series.name} but {detected_type}.contains_op(series) -> False"
    assert series in detected_type, message


def _traverse_relation_graph(series, G, node=Generic):
    match_types = []
    for tenz_type in G.successors(node):
        if series in tenz_type:
            match_types.append(tenz_type)

    assert (
        len(match_types) < 2
    ), "types contains should be mutually exclusive {match_types}".format(
        match_types=match_types
    )
    if len(match_types) == 1:
        return _traverse_relation_graph(series, G, match_types[0])
    else:
        return node


# What does this test? It doesn't explicitly invoke the actual traversal code.
def test_traversal_mutex(series):
    _traverse_relation_graph(series, typeset.relation_graph)



import datetime
import os
import pathlib
import uuid
from ipaddress import IPv4Address, IPv6Address
from pathlib import PurePosixPath, PureWindowsPath
from urllib.parse import urlparse

import numpy as np
import pandas as pd

from visions.types.email_address import FQDA

base_path = os.path.abspath(os.path.dirname(__file__))


def get_series():
    test_series = [
        # Int Series
        pd.Series([1, 2, 3], name="int_series"),
        pd.Series(range(10), name="int_range"),
        pd.Series([1, 2, 3], name="Int64_int_series", dtype="Int64"),
        pd.Series([1, 2, 3, np.nan], name="Int64_int_nan_series", dtype="Int64"),
        pd.Series([1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0], name="int_series_boolean"),
        # Count
        pd.Series(np.array([1, 2, 3, 4], dtype=np.uint32), name="np_uint32"),
        pd.Series(np.array([1, 2, 3, 4], dtype="UInt32"), name="pd_uint32"),
        # Categorical
        pd.Series([1, 2, 3], name="categorical_int_series", dtype="category"),
        pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"],
                categories=["A", "B", "C"],
                ordered=False,
            ),
            name="categorical_char",
        ),
        pd.Series([1.0, 2.0, 3.1], dtype="category", name="categorical_float_series"),
        pd.Series(
            ["Georgia", "Sam"], dtype="category", name="categorical_string_series"
        ),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1)],
            name="categorical_complex_series",
            dtype="category",
        ),
        # Ordinal
        pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"], categories=["A", "B", "C"], ordered=True
            ),
            name="ordinal",
        ),
        # Float Series
        pd.Series([1.0, 2.1, 3.0], name="float_series"),
        pd.Series([1.0, 2.5, np.nan], name="float_nan_series"),
        pd.Series([1.0, 2.0, 3.0, 4.0], name="float_series2"),
        pd.Series(np.array([1.2, 2, 3, 4], dtype=np.float), name="float_series3"),
        pd.Series([1, 2, 3.05, 4], dtype=np.float, name="float_series4"),
        pd.Series([np.nan, 1.2], name="float_series5"),
        pd.Series([np.nan, 1.1], dtype=np.single, name="float_series6"),
        pd.Series([np.inf, np.NINF, np.PINF, 1000000.0, 5.5], name="float_with_inf"),
        pd.Series([np.inf, np.NINF, np.Infinity, np.PINF], name="inf_series"),
        pd.Series([1, 2, np.nan], name="int_nan_series"),
        # Nan Series
        pd.Series([np.nan], name="nan_series"),
        pd.Series([np.nan, np.nan, np.nan, np.nan], name="nan_series_2"),
        # String Series
        pd.Series(["Patty", "Valentine"], name="string_series"),
        pd.Series(["mack", "the", "finger"], name="string_unicode_series"),
        pd.Series(
            np.array(["upper", "hall"], dtype=np.unicode_),
            name="string_np_unicode_series",
        ),
        pd.Series(["1.0", "2.0", np.nan], name="string_num_nan"),
        pd.Series(["1,000.0", "2.1", np.nan], name="string_with_sep_num_nan"),
        pd.Series(["1.0", "2.0", "3.0"], name="string_num"),
        pd.Series(["1.0", "45.67", np.nan], name="string_flt_nan"),
        pd.Series(["1.0", "45.67", "3.5"], name="string_flt"),
        pd.Series(
            [
                "I was only robbing the register,",
                "I hope you understand",
                "One of us had better call up the cops",
                "In the hot New Jersey night",
                np.nan,
            ],
            name="string_str_nan",
        ),
        pd.Series(["True", "False", None], name="string_bool_nan"),
        pd.Series(range(20), name="int_str_range").astype("str"),
        pd.Series(
            [
                "http://www.cwi.nl:80/%7Eguido/Python.html",
                "https://github.com/dylan-profiling/hurricane",
            ],
            name="str_url",
        ),
        pd.Series(
            [r"C:\\home\\user\\file.txt", r"C:\\home\\user\\test2.txt"],
            name="path_series_windows_str",
        ),
        pd.Series(
            [r"/home/user/file.txt", r"/home/user/test2.txt"],
            name="path_series_linux_str",
        ),
        pd.Series(["0011", "12"], name="str_int_leading_zeros"),
        pd.Series(["0.0", "0.04", "0"], name="str_float_non_leading_zeros"),
        pd.Series(["0.0", "0.000", "0", "2"], name="str_int_zeros"),
        # Bool Series
        pd.Series([True, False], name="bool_series"),
        pd.Series([True, False, None], name="bool_nan_series"),
        pd.Series([True, False, None], name="nullable_bool_series", dtype="Bool"),
        pd.Series([True, False, False, True], name="bool_series2", dtype=bool),
        pd.Series([True, False, False, True], name="bool_series2", dtype=bool),
        pd.Series(np.array([1, 0, 0, 1], dtype=np.bool), name="bool_series3"),
        # Complex Series
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1)],
            name="complex_series",
        ),
        pd.Series(
            [
                np.complex(0, 0),
                np.complex(1, 2),
                np.complex(3, -1),
                np.complex(np.nan, np.nan),
            ],
            name="complex_series_nan",
        ),
        pd.Series(["(1+1j)", "(2+2j)", "(10+100j)"], name="str_complex"),
        pd.Series(["(1+1j)", "(2+2j)", "(10+100j)", "NaN"], name="str_complex_nan"),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1), np.nan],
            name="complex_series_nan_2",
        ),
        pd.Series(
            [complex(0, 0), complex(1, 2), complex(3, -1), np.nan],
            name="complex_series_py_nan",
        ),
        pd.Series(
            [complex(0, 0), complex(1, 2), complex(3, -1)], name="complex_series_py"
        ),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 0), np.complex(3, 0), np.complex(-1, 0)],
            name="complex_series_float",
        ),
        # Datetime Series
        pd.Series(["1937-05-06", "20/4/2014"], name="string_date"),
        pd.Series(["1941-05-24", "13/10/2016"], name="timestamp_string_series"),
        pd.to_datetime(
            pd.Series(
                [datetime.datetime(2017, 3, 5, 12, 2), datetime.datetime(2019, 12, 4)],
                name="timestamp_series",
            )
        ),
        pd.to_datetime(
            pd.Series(
                [
                    datetime.datetime(2017, 3, 5),
                    datetime.datetime(2019, 12, 4, 3, 2, 0),
                    pd.NaT,
                ],
                name="timestamp_series_nat",
            )
        ),
        pd.to_datetime(
            pd.Series(
                [datetime.datetime(2017, 3, 5), datetime.datetime(2019, 12, 4), pd.NaT],
                name="date_series_nat",
            )
        ),
        pd.Series(
            pd.date_range(
                start="2013-05-18 12:00:01",
                periods=2,
                freq="H",
                tz="Europe/Brussels",
                name="timestamp_aware_series",
            )
        ),
        pd.to_datetime(
            pd.Series(
                [
                    datetime.date(2011, 1, 1),
                    datetime.date(2012, 1, 2),
                    datetime.date(2013, 1, 1),
                ],
                name="datetime",
            )
        ),
        # Date series
        pd.Series(
            [
                datetime.date(2011, 1, 1),
                datetime.date(2012, 1, 2),
                datetime.date(2013, 1, 1),
            ],
            name="date",
        ),
        # Time series
        pd.Series(
            [
                datetime.time(8, 43, 12),
                datetime.time(9, 43, 12),
                datetime.time(10, 43, 12),
            ],
            name="time",
        ),
        # http://pandas-docs.github.io/pandas-docs-travis/user_guide/timeseries.html#timestamp-limitations
        # pd.to_datetime(
        #     pd.Series(
        #         [
        #             datetime.datetime(year=1, month=1, day=1, hour=8, minute=43, second=12),
        #             datetime.datetime(year=1, month=1, day=1, hour=9, minute=43, second=12),
        #             datetime.datetime(
        #                 year=1, month=1, day=1, hour=10, minute=43, second=12
        #             ),
        #         ],
        #         name="datetime_to_time",
        #     )
        # ),
        # Timedelta Series
        pd.Series([pd.Timedelta(days=i) for i in range(3)], name="timedelta_series"),
        pd.Series(
            [pd.Timedelta(days=i) for i in range(3)] + [pd.NaT],
            name="timedelta_series_nat",
        ),
        pd.Series(
            [
                pd.Timedelta("1 days 00:03:43"),
                pd.Timedelta("5 days 12:33:57"),
                pd.Timedelta("0 days 01:25:07"),
                pd.Timedelta("-2 days 13:46:56"),
                pd.Timedelta("1 days 23:49:25"),
            ],
            name="timedelta_negative",
        ),
        # Path Series
        pd.Series(
            [
                PurePosixPath("/home/user/file.txt"),
                PurePosixPath("/home/user/test2.txt"),
            ],
            name="path_series_linux",
        ),
        pd.Series(
            [
                PurePosixPath("/home/user/file.txt"),
                PurePosixPath("/home/user/test2.txt"),
                None,
            ],
            name="path_series_linux_missing",
        ),
        pd.Series(
            [
                PureWindowsPath("C:\\home\\user\\file.txt"),
                PureWindowsPath("C:\\home\\user\\test2.txt"),
            ],
            name="path_series_windows",
        ),
        # Url Series
        pd.Series(
            [
                urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
                urlparse("https://github.com/dylan-profiling/hurricane"),
            ],
            name="url_series",
        ),
        pd.Series(
            [
                urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
                urlparse("https://github.com/dylan-profiling/hurricane"),
                np.nan,
            ],
            name="url_nan_series",
        ),
        pd.Series(
            [
                urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
                urlparse("https://github.com/dylan-profiling/hurricane"),
                None,
            ],
            name="url_none_series",
        ),
        # UUID Series
        pd.Series(
            [
                uuid.UUID("0b8a22ca-80ad-4df5-85ac-fa49c44b7ede"),
                uuid.UUID("aaa381d6-8442-4f63-88c8-7c900e9a23c6"),
                uuid.UUID("00000000-0000-0000-0000-000000000000"),
            ],
            name="uuid_series",
        ),
        pd.Series(
            [
                uuid.UUID("0b8a22ca-80ad-4df5-85ac-fa49c44b7ede"),
                uuid.UUID("aaa381d6-8442-4f63-88c8-7c900e9a23c6"),
                uuid.UUID("00000000-0000-0000-0000-000000000000"),
                None,
            ],
            name="uuid_series_missing",
        ),
        pd.Series(
            [
                "0b8a22ca-80ad-4df5-85ac-fa49c44b7ede",
                "aaa381d6-8442-4f63-88c8-7c900e9a23c6",
                "00000000-0000-0000-0000-000000000000",
            ],
            name="uuid_series_str",
        ),
        # Object Series
        pd.Series([[1, ""], [2, "Rubin"], [3, "Carter"]], name="mixed_list[str,int]"),
        pd.Series(
            [{"why": "did you"}, {"bring him": "in for he"}, {"aint": "the guy"}],
            name="mixed_dict",
        ),
        pd.Series(
            [pd.to_datetime, pd.to_timedelta, pd.read_json, pd.to_pickle],
            name="callable",
        ),
        pd.Series([pd, np], name="module"),
        pd.Series(["1.1", "2"], name="textual_float"),
        pd.Series(["1.1", "2", "NAN"], name="textual_float_nan"),
        # Object (Mixed, https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.types.infer_dtype.html)
        pd.Series(["a", 1], name="mixed_integer"),
        pd.Series([True, False, np.nan], name="mixed"),
        pd.Series([[True], [False], [False]], name="mixed_list"),
        pd.Series([[1, ""], [2, "Rubin"], [3, "Carter"]], name="mixed_list[str,int]"),
        pd.Series(
            [{"why": "did you"}, {"bring him": "in for he"}, {"aint": "the guy"}],
            name="mixed_dict",
        ),
        # IP
        pd.Series([IPv4Address("127.0.0.1"), IPv4Address("127.0.0.1")], name="ip"),
        pd.Series(["127.0.0.1", "127.0.0.1"], name="ip_str"),
        # Empty
        pd.Series([], name="empty"),
        pd.Series([], name="empty_float", dtype=float),
        pd.Series([], name="empty_int64", dtype="Int64"),
        pd.Series([], name="empty_object", dtype="object"),
        pd.Series([], name="empty_bool", dtype=bool),
        # IP
        pd.Series([IPv4Address("127.0.0.1"), IPv4Address("127.0.0.1")], name="ip"),
        pd.Series(
            [IPv4Address("127.0.0.1"), None, IPv4Address("127.0.0.1")],
            name="ip_missing",
        ),
        pd.Series(
            [IPv6Address("0:0:0:0:0:0:0:1"), IPv4Address("127.0.0.1")],
            name="ip_mixed_v4andv6",
        ),
        pd.Series(["127.0.0.1", "127.0.0.1"], name="ip_str"),
        # File
        pd.Series(
            [
                pathlib.Path(os.path.join(base_path, "series.py")).absolute(),
                pathlib.Path(os.path.join(base_path, "__init__.py")).absolute(),
                pathlib.Path(os.path.join(base_path, "utils.py")).absolute(),
            ],
            name="file_test_py",
        ),
        pd.Series(
            [
                pathlib.Path(os.path.join(base_path, "..", "py.typed")).absolute(),
                pathlib.Path(
                    os.path.join(
                        base_path, "..", "visualisation", "circular_packing.html"
                    )
                ).absolute(),
                pathlib.Path(os.path.join(base_path, "series.py")).absolute(),
            ],
            name="file_mixed_ext",
        ),
        pd.Series(
            [
                pathlib.Path(os.path.join(base_path, "series.py")).absolute(),
                None,
                pathlib.Path(os.path.join(base_path, "__init__.py")).absolute(),
                None,
                pathlib.Path(os.path.join(base_path, "utils.py")).absolute(),
            ],
            name="file_test_py_missing",
        ),
        # Image
        pd.Series(
            [
                pathlib.Path(
                    os.path.join(
                        base_path,
                        "../visualisation/typesets/typeset_complete.png",
                    )
                ).absolute(),
                pathlib.Path(
                    os.path.join(
                        base_path,
                        r"../visualisation/typesets/typeset_standard.png",
                    )
                ).absolute(),
                pathlib.Path(
                    os.path.join(
                        base_path,
                        r"../visualisation/typesets/typeset_geometry.png",
                    )
                ).absolute(),
            ],
            name="image_png",
        ),
        pd.Series(
            [
                pathlib.Path(
                    os.path.join(
                        base_path,
                        r"../visualisation/typesets/typeset_complete.png",
                    )
                ).absolute(),
                pathlib.Path(
                    os.path.join(
                        base_path,
                        r"../visualisation/typesets/typeset_standard.png",
                    )
                ).absolute(),
                None,
                pathlib.Path(
                    os.path.join(
                        base_path,
                        r"../visualisation/typesets/typeset_geometry.png",
                    )
                ).absolute(),
                None,
            ],
            name="image_png_missing",
        ),
        # Email
        pd.Series(
            [FQDA("test", "example.com"), FQDA("info", "example.eu")],
            name="email_address",
        ),
        pd.Series(
            [FQDA("test", "example.com"), FQDA("info", "example.eu"), None],
            name="email_address_missing",
        ),
        pd.Series(["test@example.com", "info@example.eu"], name="email_address_str"),
    ]

    if int(pd.__version__[0]) >= 1:
        pandas_1_series = [
            pd.Series(
                ["Patty", "Valentine"], dtype="string", name="string_dtype_series"
            )
        ]
        test_series.extend(pandas_1_series)

    return test_series


def get_geometry_series():
    from shapely import wkt

    series = [
        pd.Series(
            ["POINT (-92 42)", "POINT (-92 42.1)", "POINT (-92 42.2)"],
            name="geometry_string_series",
        ),
        pd.Series(
            [
                wkt.loads("POINT (-92 42)"),
                wkt.loads("POINT (-92 42.1)"),
                wkt.loads("POINT (-92 42.2)"),
            ],
            name="geometry_series",
        ),
        pd.Series(
            [
                wkt.loads("POINT (-92 42)"),
                wkt.loads("POINT (-92 42.1)"),
                wkt.loads("POINT (-92 42.2)"),
                None,
            ],
            name="geometry_series_missing",
        ),
    ]
    return series



from typing import Optional

import attr
import pandas as pd


def func_repr(func):
    return func.__name__ if hasattr(func, "__name__") else str("lambda")


def identity_relation(series: pd.Series, state: dict) -> pd.Series:
    return series


@attr.s(frozen=True)
class TypeRelation:
    """Relationship encoder between implementations of :class:`visions.types.type.VisionsBaseType`

    Defines a one to one relationship between two :class:`visions.types.type.VisionsBaseType` implementations,
    A and B, with respect to an underlying data series. In order to define a relationship we need
    two methods:

        - **is_relationship**, determines whether a series of type B can be alternatively represented as type A.
        - **transform**, provides a mechanism to convert the series from B -> A.

    For example, the series `pd.Series([1.0, 2.0, 3.0])` is encoded as a sequence of
    floats but in reality they are all integers.

    Examples:
        >>> from visions.types import Integer, Float
        >>> x = pd.Series([1.0, 2.0, 3.0])
        >>> state = dict()
        >>> relation = TypeRelation(Integer, Float)
        >>> relation.is_relation(x, state)
        True

        >>> relation.transform(x, state)
        pd.Series([1, 2, 3])
    """

    type = attr.ib()
    related_type = attr.ib()
    inferential = attr.ib()
    transformer = attr.ib(repr=func_repr)
    relationship = attr.ib(default=lambda x, y: False, repr=func_repr)

    def is_relation(self, series: pd.Series, state: Optional[dict] = None) -> bool:
        if state is None:
            state = {}
        return self.relationship(series, state)

    def transform(self, series: pd.Series, state: Optional[dict] = None) -> pd.Series:
        if state is None:
            state = {}
        return self.transformer(series, state)

    def __str__(self):
        return f"{self.related_type}->{self.type}"


@attr.s(frozen=True)
class IdentityRelation(TypeRelation):
    relationship = attr.ib(repr=func_repr)
    transformer = attr.ib(default=identity_relation, repr=func_repr)
    inferential = attr.ib(default=False)

    @relationship.default
    def make_relationship(self):
        return self.type.__contains__


@attr.s(frozen=True)
class InferenceRelation(TypeRelation):
    relationship = attr.ib(repr=func_repr)
    inferential = attr.ib(default=True)

    @relationship.default
    def make_relationship(self):
        return self.type.__contains__



from visions import String
from visions.relations.relations import InferenceRelation


def string_to_categorical_distinct_count(cls) -> InferenceRelation:
    """Convert string to categorical when it has fewer than 50% unique values.

    Returns:
        relation
    """
    # TODO: only when not any other string relation (either exclude others or have ordering and evaluate last)
    return InferenceRelation(
        relationship=lambda s, state: s.nunique() / len(s) < 0.5,
        transformer=lambda s: s.astype("category"),
        related_type=String,
        type=cls,
    )



import pandas as pd


def dataframe_summary(df: pd.DataFrame) -> dict:
    """Summarization for a DataFrame

    Args:
        df: a DataFrame object to summarize

    Returns:
        Summary of the DataFrame with `n_observations`, `n_variables` and `memory_size`.
    """
    return {
        "n_observations": df.shape[0],
        "n_variables": df.shape[1],
        "memory_size": df.memory_usage(deep=True).sum(),
    }



from visions.application.summaries.series import (
    base_summary,
    category_summary,
    file_summary,
    image_summary,
    missing_summary,
    numerical_basic_summary,
    numerical_summary,
    path_summary,
    range_summary,
    text_summary,
    unique_summary,
    unique_summary_complex,
    url_summary,
    zero_summary,
)
from visions.application.summaries.summary import Summary
from visions.types import *
from visions.typesets import CompleteSet


class CompleteSummary(Summary):
    def __init__(self) -> None:
        type_summary_ops = {
            Boolean: [],
            Categorical: [category_summary, unique_summary],
            Complex: [numerical_basic_summary, unique_summary_complex],
            DateTime: [range_summary, unique_summary],
            Date: [],
            File: [file_summary, path_summary, text_summary],
            Float: [numerical_summary, zero_summary, unique_summary],
            Geometry: [],
            Image: [image_summary],
            Integer: [numerical_summary, zero_summary, unique_summary],
            Object: [unique_summary],
            Path: [path_summary, text_summary],
            String: [text_summary, unique_summary],
            Time: [],
            TimeDelta: [],
            UUID: [],
            URL: [url_summary, unique_summary],
            Generic: [base_summary, missing_summary],
        }
        super().__init__(type_summary_ops, CompleteSet())



from typing import Any, Dict, List, Type

import networkx as nx
import pandas as pd

from visions.application.summaries.frame.dataframe_series_summary import (
    dataframe_series_summary,
)
from visions.application.summaries.frame.dataframe_summary import dataframe_summary
from visions.application.summaries.frame.dataframe_type_summary import (
    dataframe_type_summary,
)
from visions.types import VisionsBaseType
from visions.utils.graph import output_graph


class Summary(object):
    def __init__(self, summary_ops, typeset):
        """

        Args:
            summary_ops:
            typeset:
        """
        self.typeset = typeset
        if summary_ops is None:
            summary_ops = {}

        if not all(
            issubclass(base_type, VisionsBaseType) for base_type in summary_ops.keys()
        ):
            raise TypeError("Summaries must be mapped on a type!")

        self.summary_ops = summary_ops

    def summarize_frame(
        self, df: pd.DataFrame, series_summary: dict, series_types: dict
    ):
        """Summarize a DataFrame based on the DataFrame object and the summaries of individual series

        Args:
            df: the DataFrame object
            series_summary: mapping from column name to the individual summaries
            series_types: mapping from column name to the series' type

        Returns:
            A summary of the DataFrame
        """
        return {
            **dataframe_summary(df),
            **dataframe_type_summary(series_types),
            **dataframe_series_summary(series_summary),
        }

    def summarize_series(
        self, series: pd.Series, summary_type: Type[VisionsBaseType]
    ) -> Dict[str, Any]:
        """

        Args:
            series:
            summary_type:

        Returns:

        """
        summary: Dict[str, Any] = {}

        G = self.typeset.base_graph.copy()

        done: List[Any] = []
        for base_type, summary_ops in self.summary_ops.items():
            if base_type not in done and nx.has_path(G, base_type, summary_type):
                for op in summary_ops:
                    summary.update(op(series))
                done.append(base_type)

        return summary

    def summarize(self, df: pd.DataFrame, types: dict) -> dict:
        """

        Args:
            df:
            types:

        Returns:

        """
        series_summary = {
            col: self.summarize_series(df[col], types[col]) for col in df.columns
        }
        frame_summary = self.summarize_frame(df, series_summary, types)
        return {"types": types, "series": series_summary, "frame": frame_summary}

    def plot(self, file_name, type_specific=None):
        """

        Args:
            file_name:
            type_specific:

        Returns:

        """
        G = self.typeset.base_graph.copy()
        G.graph["node"] = {"shape": "box", "color": "red"}

        included_nodes = G.nodes
        if type_specific is not None:
            included_nodes = nx.ancestors(G, type_specific)
            included_nodes.add(type_specific)
            G.remove_nodes_from(G.nodes - included_nodes)

        G.add_node("summary", shape="note")
        for base_type, summary_ops in self.summary_ops.items():
            if len(summary_ops) > 0 and base_type in included_nodes:
                G.add_edge(
                    str(base_type),
                    "summary",
                    label="\n".join([str(op.__name__) for op in summary_ops]),
                )

        output_graph(G, file_name)



import datetime

import numpy as np
import pandas as pd

from visions.functional import (
    cast_to_detected,
    cast_to_inferred,
    detect_type,
    infer_type,
)
from visions.types import Complex, DateTime, Integer, String
from visions.typesets import CompleteSet, StandardSet


def test_type_inference_frame():
    # Create a DataFrame with various string columns
    df = pd.DataFrame(
        {
            "latin": ["orange", "apple", "pear"],
            "cyrillic": ["Кириллица", "гласность", "демократија"],
            "mixed": ["Кириллица", "soep", "демократија"],
            "burmese": ["ရေကြီးခြင်း", "စက်သင်ယူမှု", "ဉာဏ်ရည်တု"],
            "digits": ["1234", "121223", "12312"],
            "specials": ["$", "%^&*(", "!!!~``"],
            "whitespace": ["\t", "\n", " "],
            "jiddisch": ["רעכט צו לינקס", "שאָסיי 61", "פּיצאַ איז אָנגענעם"],
            "arabic": ["بوب ديلان", "باتي فالنتين", "السيد الدف الرجل"],
            "playing_cards": ["🂶", "🃁", "🂻"],
        }
    )

    # Initialize the typeset
    typeset = CompleteSet()

    # Infer the column type
    types = infer_type(df, typeset)
    assert types == {
        "latin": String,
        "cyrillic": String,
        "mixed": String,
        "burmese": String,
        "digits": Integer,
        "specials": String,
        "whitespace": String,
        "jiddisch": String,
        "arabic": String,
        "playing_cards": String,
    }


def test_type_inference_series():
    string_series = pd.Series(["(12.0+10.0j)", "(-4.0+6.2j)", "(8.0+2.0j)"])

    typeset = StandardSet()
    detected_type = infer_type(string_series, typeset)
    assert detected_type == Complex


def test_type_cast_infer_frame():
    df = pd.DataFrame(
        {
            "latin": ["orange", "apple", "pear"],
            "cyrillic": ["Кириллица", "гласность", "демократија"],
            "mixed": ["Кириллица", "soep", "демократија"],
            "burmese": ["ရေကြီးခြင်း", "စက်သင်ယူမှု", "ဉာဏ်ရည်တု"],
            "digits": ["1234", "121223", "12312"],
            "specials": ["$", "%^&*(", "!!!~``"],
            "whitespace": ["\t", "\n", " "],
            "jiddisch": ["רעכט צו לינקס", "שאָסיי 61", "פּיצאַ איז אָנגענעם"],
            "arabic": ["بوب ديلان", "باتي فالنتين", "السيد الدف الرجل"],
            "playing_cards": ["🂶", "🃁", "🂻"],
        }
    )

    typeset = CompleteSet()
    new_df = cast_to_inferred(df, typeset)
    print(new_df)
    assert new_df["digits"].iloc[1] - 3 == 121220
    assert new_df["latin"].iloc[1] + "1" == "apple1"


def test_type_cast_infer_series():
    string_series = pd.Series(["(12.0+10.0j)", "(-4.0+6.2j)", "(8.0+2.0j)"])

    typeset = StandardSet()
    new_series = cast_to_inferred(string_series, typeset)
    assert new_series.iloc[1].real == -4.0


def test_type_cast_detect_series():
    string_series = pd.Series(["(12.0+10.0j)", "(-4.0+6.2j)", "(8.0+2.0j)"])

    typeset = StandardSet()
    new_series = cast_to_detected(string_series, typeset)
    assert new_series.iloc[1] == "(-4.0+6.2j)"


def test_type_cast_detect_frame():
    df = pd.DataFrame(
        {
            "latin": ["orange", "apple", "pear"],
            "cyrillic": ["Кириллица", "гласность", "демократија"],
            "mixed": ["Кириллица", "soep", "демократија"],
            "burmese": ["ရေကြီးခြင်း", "စက်သင်ယူမှု", "ဉာဏ်ရည်တု"],
            "digits": ["1234", "121223", "12312"],
            "specials": ["$", "%^&*(", "!!!~``"],
            "whitespace": ["\t", "\n", " "],
            "jiddisch": ["רעכט צו לינקס", "שאָסיי 61", "פּיצאַ איז אָנגענעם"],
            "arabic": ["بوب ديلان", "باتي فالنتين", "السيد الدف الرجل"],
            "playing_cards": ["🂶", "🃁", "🂻"],
        }
    )

    typeset = CompleteSet()
    new_df = cast_to_detected(df, typeset)
    assert new_df["digits"].iloc[1] == "121223"
    assert new_df["latin"].iloc[1] + "1" == "apple1"


def test_type_detect_frame():
    # Create a DataFrame with various string columns
    df = pd.DataFrame(
        {
            "latin": ["orange", "apple", "pear"],
            "cyrillic": ["Кириллица", "гласность", "демократија"],
            "mixed": ["Кириллица", "soep", "демократија"],
            "burmese": ["ရေကြီးခြင်း", "စက်သင်ယူမှု", "ဉာဏ်ရည်တု"],
            "digits": ["01234", "121223", "12312"],
            "specials": ["$", "%^&*(", "!!!~``"],
            "whitespace": ["\t", "\n", " "],
            "jiddisch": ["רעכט צו לינקס", "שאָסיי 61", "פּיצאַ איז אָנגענעם"],
            "arabic": ["بوب ديلان", "باتي فالنتين", "السيد الدف الرجل"],
            "playing_cards": ["🂶", "🃁", "🂻"],
        }
    )

    # Initialize the typeset
    typeset = CompleteSet()

    # Infer the column type
    types = detect_type(df, typeset)
    assert types == {
        "latin": String,
        "cyrillic": String,
        "mixed": String,
        "burmese": String,
        "digits": String,
        "specials": String,
        "whitespace": String,
        "jiddisch": String,
        "arabic": String,
        "playing_cards": String,
    }


def test_type_detect_series():
    datetime_series = pd.Series(
        [
            datetime.datetime(2010, 1, 1),
            datetime.datetime(2010, 8, 2),
            datetime.datetime(2011, 2, 1),
            np.datetime64("NaT"),
        ]
    )

    typeset = StandardSet()
    detected_type = detect_type(datetime_series, typeset)
    assert detected_type == DateTime



"""
A selection of testing utilities for visions.
"""
import functools
from typing import Callable, Dict, List, Optional, Type, Union

import pandas as pd

from visions.utils.series_utils import func_nullable_series_contains


def option_coercion_evaluator(
    method: Callable, extra_errors: Optional[List[Type[Exception]]] = None
) -> Callable:
    """A coercion test evaluator

    Evaluates a coercion method and optionally returns the coerced series.

    Args:
        method: A method coercing a Series to another Series.

    Returns:
        The coerced series if the coercion succeeds otherwise None.
    """

    error_list = [ValueError, TypeError, AttributeError]
    if extra_errors:
        error_list.extend(extra_errors)

    @functools.wraps(method)
    def f(series: pd.Series) -> Optional[pd.Series]:
        try:
            return method(series)
        except tuple(error_list):
            return None

    return f


def coercion_test(
    method: Callable, extra_errors: Optional[List[Type[Exception]]] = None
) -> Callable:
    """A coercion test generator

    Creates a coercion test based on a provided coercion method.

    Args:
        method: A method coercing a Series to another type.

    Returns:
        Whether the coercion failed or was successful.

    """
    # Returns True or False if the coercion succeeds
    tester = option_coercion_evaluator(method, extra_errors)

    @functools.wraps(method)
    def f(series: pd.Series) -> bool:
        result = tester(series)
        return True if result is not None else False

    return f


def coercion_true_test(
    method: Callable, extra_errors: Optional[List[Type[Exception]]] = None
) -> Callable:
    """A coercion equality test generator

    Creates a coercion test based on a provided coercion method which also enforces
    equality constraints on the output. This is useful when you want to change the
    data type of a series without necessarily changing the data, for example,
    when converting an integer to a float.

    Args:
        method: A method coercing a Series to another type.

    Returns:
        Whether the coercion failed or was successful.
    """
    tester = option_coercion_evaluator(method, extra_errors)

    @functools.wraps(tester)
    def f(series: pd.Series) -> bool:
        result = tester(series)
        return False if result is None else series.all()

    return f


def coercion_equality_test(method: Callable) -> Callable:
    """A coercion equality test generator

    Creates a coercion test based on a provided coercion method which also enforces
    equality constraints on the output. This is useful when you want to change the
    data type of a series without necessarily changing the data, for example,
    when converting an integer to a float.

    Args:
        method: A method coercing a Series to another type.

    Returns:
        Whether the coercion failed or was successful.
    """
    tester = option_coercion_evaluator(method)

    @functools.wraps(tester)
    def f(series: pd.Series) -> bool:
        result = tester(series)
        return False if result is None else series.eq(result).all()

    return f


def coercion_single_map_test(mapping: List[Dict]) -> Callable:
    @func_nullable_series_contains
    def f(series: pd.Series, state: dict = {}) -> bool:
        return any(series.isin(list(single_map.keys())).all() for single_map in mapping)

    return f


def coercion_multi_map_test(mapping: Dict) -> Callable:
    @func_nullable_series_contains
    def f(series: pd.Series, state: dict = {}) -> bool:
        return series.isin(list(mapping.keys())).all()

    return f


def coercion_map_test(mapping: Union[List[Dict], Dict]) -> Callable:
    """Create a testing function for a single mapping or a list of mappings.

    Args:
        mapping: A dict with a mapping or a list of dicts

    Returns:
        Callable that checks if a series consists of the mappable values

    Examples:
        >>> coercion_map_test({"Yes": True, "No": False})
        >>> coercion_map_test(
        >>>     [
        >>>         {"Yes": True, "No": False},
        >>>         {"Y": True, "N": False},
        >>>     ]
        >>> )
    """

    if isinstance(mapping, list):
        f = coercion_single_map_test(mapping)
    elif isinstance(mapping, dict):
        f = coercion_multi_map_test(mapping)
    else:
        raise ValueError("Mapping should be dict or list of dicts")
    return f


def coercion_map(mapping: Union[List[Dict], Dict]) -> Callable:
    """Maps a series given a mapping

    Args:
        mapping: a dict to map, or a list of dicts.

    Returns:
        A callable that maps the series.
    """
    if type(mapping) == list:
        mapping = {k: v for d in mapping for k, v in d.items()}

    elif type(mapping) != dict:
        raise ValueError("Mapping should be dict or list of dicts")

    def f(series: pd.Series) -> pd.Series:
        return series.map(mapping)

    return f



import pandas as pd


def zero_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {"n_zeros": (series == 0).sum()}
    return summary



import functools
from typing import Callable

import pandas as pd


def nullable_series_contains(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def inner(cls, series: pd.Series, state: dict = {}, *args, **kwargs) -> bool:
        if series.hasnans:
            series = series.dropna()
            if series.empty:
                return False

        return fn(cls, series, state, *args, **kwargs)

    return inner


def func_nullable_series_contains(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def inner(series: pd.Series, state={}, *args, **kwargs) -> bool:
        if series.hasnans:
            series = series.dropna()
            if series.empty:
                return False

        return fn(series, state, *args, **kwargs)

    return inner


def _contains_instance_attrs(
    series, is_method, class_name, attrs: list, sample_size=1
) -> bool:
    # TODO: user configurable .head or .sample
    # TODO: performance testing for series[0], series.iloc[0], series.head, series.sample
    if not all(is_method(x, class_name) for x in series.head(sample_size)):
        return False

    try:
        return all(all(hasattr(x, attr) for attr in attrs) for x in series)
    except AttributeError:
        return False


def class_name_attrs(series, class_name, attrs: list, sample_size=1) -> bool:
    def func(instance, class_name):
        return instance.__class__.__name__ == class_name.__name__

    return _contains_instance_attrs(series, func, class_name, attrs, sample_size)


def isinstance_attrs(series, class_name, attrs: list, sample_size=1) -> bool:
    return _contains_instance_attrs(series, isinstance, class_name, attrs, sample_size)



import pandas as pd


def unique_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {"n_unique": series.nunique(), "is_unique": series.is_unique}
    return summary


def unique_summary_complex(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Notes:
        Until complex bug is fixed:
        - https://github.com/pandas-dev/pandas/issues/17927
        - https://github.com/pandas-dev/pandas/pull/27599

    Returns:

    """

    n_unique = len(set(series[series.notna()].values))
    summary = {"n_unique": n_unique, "is_unique": series.is_unique}
    return summary



import pandas as pd

from visions.application.summaries import CompleteSummary
from visions.functional import detect_type
from visions.typesets import CompleteSet

# Create a DataFrame with various string columns
df = pd.DataFrame(
    {
        "latin": ["orange", "apple", "pear"],
        "cyrillic": ["Кириллица", "гласность", "демократија"],
        "mixed": ["Кириллица", "soep", "демократија"],
        "burmese": ["ရေကြီးခြင်း", "စက်သင်ယူမှု", "ဉာဏ်ရည်တု"],
        "digits": ["01234", "121223", "123123"],
        "specials": ["$", "%^&*(", "!!!~``"],
        "whitespace": ["\t", "\n", " "],
        "jiddisch": ["רעכט צו לינקס", "שאָסיי 61", "פּיצאַ איז אָנגענעם"],
        "arabic": ["بوب ديلان", "باتي فالنتين", "السيد الدف الرجل"],
        "playing_cards": ["🂶", "🃁", "🂻"],
    }
)

# Initialize the typeset
typeset = CompleteSet()

# Infer the column type
types = detect_type(df, typeset)

# Generate a summary
summarizer = CompleteSummary()
summary = summarizer.summarize(df, types)

print(
    "| {h1: <15}| {h2: <17}| {h3: <84}| {h4: <25}|".format(
        h1="Column", h2="Scripts", h3="Categories", h4="Blocks"
    )
)
print("{e:-<17}+{e:-<18}+{e:-<85}+{e:-<26}+".format(e=""))
for column, variable_summary in summary["series"].items():
    scripts = ", ".join(set(variable_summary["script_values"].values()))
    categories = ", ".join(set(variable_summary["category_alias_values"].values()))
    blocks = ", ".join(set(variable_summary["block_values"].values()))

    print(
        "| {column: <15}| {scripts: <17}| {categories: <84}| {blocks: <25}|".format(
            column=column, scripts=scripts, categories=categories, blocks=blocks
        )
    )



import pandas as pd

from visions.relations import InferenceRelation
from visions.relations.relations_utils import values_are_consecutive
from visions.types.ordinal import to_ordinal


def is_ordinal_int(s: pd.Series, state: dict) -> bool:
    initial_element = 1
    distinct_values = list(s.unique())
    return (
        initial_element in distinct_values
        and values_are_consecutive(distinct_values)
        and 2 < len(distinct_values) < 10
    )


def integer_to_ordinal(cls) -> InferenceRelation:
    from visions.types import Integer

    return InferenceRelation(
        cls, Integer, relationship=is_ordinal_int, transformer=to_ordinal
    )



import pandas as pd


def missing_summary(series: pd.Series) -> dict:
    """Counts the number of missing values

    Args:
        series: series to summarize

    Returns:
        A dict containing `na_count`.
    """
    mask = series.isna()
    summary = {"na_count": mask.values.sum()}
    return summary



import pandas as pd

import visions as v
from visions.application.summaries import CompleteSummary

string_series = pd.Series(["orange", "apple", "pear", "🂶", "🃁", "🂻"])

summarizer = CompleteSummary()
summary = summarizer.summarize_series(string_series, v.String)
print(summary)

# Output:
# {
#     "n_unique": 6,
#     "length": {1: 3, 6: 1, 5: 1, 4: 1},
#     "category_short_values": {
#         "o": "Ll",
#         "r": "Ll",
#         "a": "Ll",
#         "n": "Ll",
#         "g": "Ll",
#         "e": "Ll",
#         "p": "Ll",
#         "l": "Ll",
#         "🂶": "So",
#         "🃁": "So",
#         "🂻": "So",
#     },
#     "category_alias_values": {
#         "o": "Lowercase_Letter",
#         "r": "Lowercase_Letter",
#         "a": "Lowercase_Letter",
#         "n": "Lowercase_Letter",
#         "g": "Lowercase_Letter",
#         "e": "Lowercase_Letter",
#         "p": "Lowercase_Letter",
#         "l": "Lowercase_Letter",
#         "🂶": "Other_Symbol",
#         "🃁": "Other_Symbol",
#         "🂻": "Other_Symbol",
#     },
#     "script_values": {
#         "o": "Latin",
#         "r": "Latin",
#         "a": "Latin",
#         "n": "Latin",
#         "g": "Latin",
#         "e": "Latin",
#         "p": "Latin",
#         "l": "Latin",
#         "🂶": "Common",
#         "🃁": "Common",
#         "🂻": "Common",
#     },
#     "block_values": {
#         "o": "Basic Latin",
#         "r": "Basic Latin",
#         "a": "Basic Latin",
#         "n": "Basic Latin",
#         "g": "Basic Latin",
#         "e": "Basic Latin",
#         "p": "Basic Latin",
#         "l": "Basic Latin",
#         "🂶": "Playing Cards",
#         "🃁": "Playing Cards",
#         "🂻": "Playing Cards",
#     },
#     "block_alias_values": {
#         "o": "ASCII",
#         "r": "ASCII",
#         "a": "ASCII",
#         "n": "ASCII",
#         "g": "ASCII",
#         "e": "ASCII",
#         "p": "ASCII",
#         "l": "ASCII",
#         "🂶": "Playing Cards",
#         "🃁": "Playing Cards",
#         "🂻": "Playing Cards",
#     },
#     "frequencies": {"🃁": 1, "orange": 1, "🂶": 1, "pear": 1, "🂻": 1, "apple": 1},
#     "n_records": 6,
#     "memory_size": 593,
#     "dtype": dtype("O"),
#     "types": {"str": 6},
#     "na_count": 0,
# }



import pandas as pd


def numerical_basic_summary(series: pd.Series) -> dict:
    """Summary with basic aggregates

    Args:
        series: series to summarize

    Returns:
        A summary of aggregates of `mean`, `std`, `var`, `min`, `max` and `sum`.

    """
    aggregates = ["mean", "std", "var", "min", "max", "sum"]
    summary = series.agg(aggregates).to_dict()
    return summary



import numpy as np
import pandas as pd
import pytest

from visions.types.boolean import hasnan_bool_name


@pytest.mark.parametrize(
    "series,expected_values,expected_dtype",
    [
        (
            pd.Series([True, False], dtype=hasnan_bool_name),
            (True, False),
            hasnan_bool_name,
        ),
        (pd.Series([True, False], dtype=bool), (True, False), bool),
        (
            pd.Series([True, False, None, True, False, None], dtype=hasnan_bool_name),
            (True, False, None, True, False, None),
            hasnan_bool_name,
        ),
        (
            pd.Series([True, False, None, True, False, None], dtype=bool),
            (True, False, False, True, False, False),
            bool,
        ),
        (
            pd.Series(
                [True, False, np.nan, True, False, np.nan], dtype=hasnan_bool_name
            ),
            (True, False, None, True, False, None),
            hasnan_bool_name,
        ),
        (
            pd.Series([True, False, np.nan, True, False, np.nan], dtype=bool),
            (True, False, True, True, False, True),
            bool,
        ),
    ],
)
def test_series(series, expected_values, expected_dtype):
    value_series = pd.Series(expected_values)
    assert series[series.notna()].eq(value_series[value_series.notna()]).all()
    assert series.dtype == expected_dtype



import functools
import os
import sys
import warnings


def suppress_warnings(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return func(*args, **kwargs)

    return inner


def discard_stderr(func):
    """Shapely logs failures at a silly severity, just trying to suppress it's output on failures.
    Only known way to get rid of sys output when wkt.loads hits a bad value"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        sys.stderr = open(os.devnull, "w")
        res = func(*args, **kwargs)
        sys.stderr = sys.__stderr__
        return res

    return wrapper



from pprint import pprint

import numpy as np
import pandas as pd

import visions as v
from visions.application.summaries import CompleteSummary

datetime_series = pd.Series(
    [
        pd.datetime(2010, 1, 1),
        pd.datetime(2010, 8, 2),
        pd.datetime(2011, 2, 1),
        np.datetime64("NaT"),
    ]
)

# Generate a summary
summarizer = CompleteSummary()
summary = summarizer.summarize_series(datetime_series, v.DateTime)

pprint(summary)



import pandas as pd

from visions.relations import InferenceRelation
from visions.relations.string_to_datetime import to_datetime_year_month_day
from visions.types import Integer
from visions.utils.coercion import test_utils


def to_datetime(series: pd.Series) -> pd.Series:
    return pd.to_datetime(series)


def _to_datetime(cls, func) -> InferenceRelation:
    return InferenceRelation(
        relationship=test_utils.coercion_test(lambda s: func(s.astype(str))),
        transformer=to_datetime,
        type=cls,
        related_type=Integer,
    )


# TODO: do only convert obvious dates (20191003000000)
def integer_to_datetime(cls):
    return _to_datetime(cls, to_datetime)


def integer_to_datetime_year_month_day(cls) -> InferenceRelation:
    return _to_datetime(cls, to_datetime_year_month_day)



import numpy as np
import pandas as pd

import visions as v

print(pd.Series([True, False]) in v.Boolean)  # True
print(pd.Series([True, False]) in v.Generic)  # True
print(pd.Series([True, False]) in v.Integer)  # False
print(pd.Series([True, False, np.nan]) in v.Boolean)  # False
print(pd.Series([True, False, None], dtype="Bool") in v.Boolean)  # True
print(pd.Series([True, False, np.nan]) in v.Object)  # True
print(pd.Series([True, False, np.nan], dtype="Bool") in v.Boolean)  # True

print(pd.Series([0, 1, 2, 3, np.nan], dtype="Int64") in v.Integer)  # True
print(pd.Series([0, 1, 2, 3], dtype=int) in v.Integer)  # True
print(pd.Series([0, 1, 2, 3], dtype="Int64") in v.Integer)  # True



import pandas as pd


def base_summary(series: pd.Series) -> dict:
    """Summary for every series

    Args:
        series: series to summarize

    Returns:
        A summary of the series including `frequencies`, `n_records`, `memory_size`, `dtype` and `types`.
    """
    summary = {
        "frequencies": series.value_counts().to_dict(),
        "n_records": series.shape[0],
        "memory_size": series.memory_usage(deep=True),
        "dtype": series.dtype,
        "types": series.map(lambda x: type(x).__name__).value_counts().to_dict(),
    }

    return summary



# type: ignore
from pathlib import Path


def _copy(self, target):
    """Monkeypatch for pathlib

    Args:
        self:
        target:

    Returns:

    """
    import shutil

    assert self.is_file()
    shutil.copy(str(self), str(target))  # str() only there for Python < (3, 6)


Path.copy = _copy



from functools import partial
from pathlib import Path

import pandas as pd

from visions.application.summaries.series.numerical_summary import (
    named_aggregate_summary,
)
from visions.utils.images.image_utils import (
    extract_exif,
    hash_image,
    is_image_truncated,
    open_image,
)


def count_duplicate_hashes(image_descriptions: dict) -> int:
    """

    Args:
        image_descriptions:

    Returns:

    """
    counts = pd.Series(
        [x["hash"] for x in image_descriptions if "hash" in x]
    ).value_counts()
    return counts.sum() - len(counts)


def extract_exif_series(image_exifs: list) -> dict:
    """

    Args:
        image_exifs:

    Returns:

    """
    exif_keys = []
    exif_values: dict = {}

    for image_exif in image_exifs:
        # Extract key
        exif_keys.extend(list(image_exif.keys()))

        # Extract values per key
        for exif_key, exif_val in image_exif.items():
            if exif_key not in exif_values:
                exif_values[exif_key] = []

            exif_values[exif_key].append(exif_val)

    series = {"exif_keys": pd.Series(exif_keys).value_counts().to_dict()}

    for k, v in exif_values.items():
        series[k] = pd.Series(v).value_counts()

    return series


def extract_image_information(
    path: Path, exif: bool = False, hash: bool = False
) -> dict:
    """Extracts all image information per file, as opening files is slow

    Args:
        path: Path to the image
        exif: extract exif information
        hash: calculate hash (for duplicate detection)

    Returns:
        A dict containing image information
    """
    information: dict = {}
    image = open_image(path)
    information["opened"] = image is not None
    if image is not None:
        information["truncated"] = is_image_truncated(image)
        if not information["truncated"]:
            information["size"] = image.size
            if exif:
                information["exif"] = extract_exif(image)
            if hash:
                information["hash"] = hash_image(image)

    return information


def image_summary(series: pd.Series, exif: bool = False, hash: bool = False) -> dict:
    """

    Args:
        series: series to summarize
        exif: extract exif information
        hash: calculate hash (for duplicate detection)

    Returns:

    """

    image_information = series.apply(
        partial(extract_image_information, exif=exif, hash=hash)
    )
    summary = {
        "n_truncated": sum(
            [1 for x in image_information if "truncated" in x and x["truncated"]]
        ),
        "image_dimensions": pd.Series(
            [x["size"] for x in image_information if "size" in x],
            name="image_dimensions",
        ),
    }

    image_widths = summary["image_dimensions"].map(lambda x: x[0])
    summary.update(named_aggregate_summary(image_widths, "width"))
    image_heights = summary["image_dimensions"].map(lambda x: x[1])
    summary.update(named_aggregate_summary(image_heights, "height"))
    image_areas = image_widths * image_heights
    summary.update(named_aggregate_summary(image_areas, "area"))

    if hash:
        summary["n_duplicate_hash"] = count_duplicate_hashes(image_information)

    if exif:
        exif_series = extract_exif_series(
            [x["exif"] for x in image_information if "exif" in x]
        )
        summary["exif_keys_counts"] = exif_series["exif_keys"]
        summary["exif_data"] = exif_series

    return summary



import pandas as pd

from visions.relations.relations import InferenceRelation
from visions.relations.relations_utils import values_are_consecutive
from visions.types.ordinal import to_ordinal
from visions.utils import func_nullable_series_contains


@func_nullable_series_contains
def is_ordinal_cat(series: pd.Series, state: dict) -> bool:
    initial_element = "a"
    s = series.astype(str)
    if s.str.len().max() == 1:
        distinct_values = list(s.str.lower().unique())
        return initial_element in distinct_values and values_are_consecutive(
            list(map(ord, distinct_values))
        )
    else:
        return False


def categorical_to_ordinal(cls) -> InferenceRelation:
    from visions.types import Categorical

    return InferenceRelation(
        cls, Categorical, relationship=is_ordinal_cat, transformer=to_ordinal
    )



from functools import partial
from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.relations.string_to_bool import get_boolean_coercions
from visions.types.type import VisionsBaseType
from visions.utils import func_nullable_series_contains
from visions.utils.coercion.test_utils import coercion_map, coercion_map_test

hasnan_bool_name = "boolean" if int(pd.__version__.split(".")[0]) >= 1 else "Bool"


def to_bool(series: pd.Series, state: dict) -> pd.Series:
    dtype = hasnan_bool_name if series.hasnans else bool
    return series.astype(dtype)


@func_nullable_series_contains
def object_is_bool(series: pd.Series, state: dict) -> bool:
    bool_set = {True, False}
    try:
        ret = all(item in bool_set for item in series)
    except:
        ret = False

    return ret


def string_is_bool(series, state: dict, string_coercions):
    return coercion_map_test(string_coercions)(series.str.lower())


def string_to_bool(series, state: dict, string_coercions):
    return to_bool(coercion_map(string_coercions)(series.str.lower()), state)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, Object, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=partial(string_is_bool, string_coercions=cls.string_coercions),
            transformer=partial(string_to_bool, string_coercions=cls.string_coercions),
        ),
        InferenceRelation(
            cls, Object, relationship=object_is_bool, transformer=to_bool
        ),
    ]
    return relations


class Boolean(VisionsBaseType):
    """**Boolean** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([True, False, False, True])
        >>> x in visions.Boolean
        True

        >>> x = pd.Series([True, False, None])
        >>> x in visions.Boolean
        True
    """

    string_coercions = get_boolean_coercions("en")

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        if not pdt.is_categorical_dtype(series) and pdt.is_bool_dtype(series):
            return True

        return False

    @classmethod
    def make_string_coercion(cls, type_name, string_coercions):
        @classmethod
        def get_relations(cls):
            return _get_relations(cls)

        return type(
            "{name}[{type_name}]".format(name=cls.__name__, type_name=type_name),
            (cls,),
            {
                "string_coercions": string_coercions,
                "get_relations": get_relations,
                "contains_op": cls.contains_op,
                "make_string_coercion": cls.make_string_coercion,
            },
        )



import pandas as pd

from tests.typesets.test_complete_set import convert_map
from visions.test.series import get_series
from visions.utils.profiling import (
    profile_relation_is_relation,
    profile_relation_transform,
    profile_type,
)


def performance_report(membership=True):
    series_dict = {s.name: s for s in get_series()}
    conv_map = convert_map
    performance_list = []

    for type, series_names in conv_map.items():
        if membership:
            # True: "series in type"
            test_series = {name: series_dict[name] for name in series_names}
        else:
            # False: "series in type"
            test_series = {
                s.name: s for s in series_dict.values() if s.name not in series_names
            }

        performance_list.extend(profile_type(type, test_series))

    df = pd.DataFrame.from_records(performance_list)

    df["type"] = df["type"].astype(str)
    aggs = ["min", "max"]
    agg_labels = ["worst", "best"]
    summary_cols = ["series", "big O"]
    agg_df = df.groupby("type").agg(aggs).reset_index()[["type"] + summary_cols]
    agg_df.columns = ["_".join(col).strip("_") for col in agg_df.columns]
    colrenames = {
        f"{name}_{agg}": f"{rename} {name}"
        for name in summary_cols
        for agg, rename in zip(aggs, agg_labels)
    }
    agg_df.rename(columns=colrenames, inplace=True)
    df["normed run time"] = df["average run time"] / df["average run time"].min()
    df = df.groupby("type")["normed run time"].describe().sort_values("50%")
    df = pd.merge(df, agg_df, on="type", how="left")
    return df


def get_relation(to_type, from_type):
    return next(
        relation for relation in to_type.relations if relation.related_type is from_type
    )


def relations_is_relation_test():
    relation_tests = {
        get_relation(*conversions[0:2]): conversions[2] for conversions in convert_map
    }

    series_dict = {s.name: s for s in get_series()}
    performance_list = []
    for relation, names in relation_tests.items():
        test_series = {name: series_dict[name] for name in names}
        performance_list.extend(profile_relation_is_relation(relation, test_series))
    df = pd.DataFrame.from_records(performance_list)
    grouper = "relation"
    df[grouper] = df[grouper].astype(str)
    aggs = ["min", "max"]
    agg_labels = ["worst", "best"]
    summary_cols = ["series", "big O"]
    agg_df = df.groupby(grouper).agg(aggs).reset_index()[[grouper] + summary_cols]
    agg_df.columns = ["_".join(col).strip("_") for col in agg_df.columns]
    colrenames = {
        f"{name}_{agg}": f"{rename} {name}"
        for name in summary_cols
        for agg, rename in zip(aggs, agg_labels)
    }
    agg_df.rename(columns=colrenames, inplace=True)
    df["normed run time"] = df["average run time"] / df["average run time"].min()
    df = df.groupby(grouper)["normed run time"].describe().sort_values("50%")
    df = pd.merge(df, agg_df, on=grouper, how="left")
    return df


def relations_transform_test():
    relation_tests = {
        get_relation(*conversions[0:2]): conversions[2] for conversions in convert_map
    }

    series_dict = {s.name: s for s in get_series()}
    performance_list = []
    for relation, names in relation_tests.items():
        test_series = {name: series_dict[name] for name in names}
        performance_list.extend(profile_relation_transform(relation, test_series))
    df = pd.DataFrame.from_records(performance_list)
    grouper = "relation"
    df[grouper] = df[grouper].astype(str)
    aggs = ["min", "max"]
    agg_labels = ["worst", "best"]
    summary_cols = ["series", "big O"]
    agg_df = df.groupby(grouper).agg(aggs).reset_index()[[grouper] + summary_cols]
    agg_df.columns = ["_".join(col).strip("_") for col in agg_df.columns]
    colrenames = {
        f"{name}_{agg}": f"{rename} {name}"
        for name in summary_cols
        for agg, rename in zip(aggs, agg_labels)
    }
    agg_df.rename(columns=colrenames, inplace=True)
    df["normed run time"] = df["average run time"] / df["average run time"].min()
    df = df.groupby(grouper)["normed run time"].describe().sort_values("50%")
    df = pd.merge(df, agg_df, on=grouper, how="left")
    return df



from typing import Sequence
from urllib.parse import ParseResult, urlparse

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import (
    func_nullable_series_contains,
    isinstance_attrs,
    nullable_series_contains,
)


@func_nullable_series_contains
def string_is_url(series, state: dict) -> bool:
    try:
        return to_url(series, state).apply(lambda x: x.netloc and x.scheme).all()
    except AttributeError:
        return False


def to_url(series: pd.Series, state: dict) -> pd.Series:
    return series.apply(urlparse)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(cls, String, relationship=string_is_url, transformer=to_url),
    ]
    return relations


class URL(VisionsBaseType):
    """**Url** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import pandas as pd
        >>> from urllib.parse import urlparse
        >>> urls = ['http://www.cwi.nl:80/%7Eguido/Python.html', 'https://github.com/pandas-profiling/pandas-profiling']
        >>> x = pd.Series([urlparse(url) for url in urls])
        >>> x in visions.URL
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return isinstance_attrs(series, ParseResult, ["netloc", "scheme"])



"""
Test for issue 85
https://github.com/dylan-profiler/visions/issues/85
"""
import pytest

from visions import VisionsTypeset
from visions.types import *


class ExampleTypeSet(VisionsTypeset):
    """Typeset for testing: String is not included as type, however there are
    types with inference relations from string"""

    def __init__(self):
        types = {
            Generic,
            Categorical,
            Boolean,
            Float,
            DateTime,
            URL,
            Complex,
            Path,
            File,
            Image,
            Integer,
            Object,
        }
        super().__init__(types)


def test_typeset_external_type():
    """When a typeset contains at least one type that has a relation from a type that is not in the typeset:
    - the user should be warned
    - the typeset should not include that type
    - that type should not be present in the graph when printed

    """

    # Expecting warnings to match
    with pytest.warns(UserWarning, match="but String was not included in") as record:
        pp_typeset = ExampleTypeSet()

    # Expecting 6 warnings
    assert len(record) == 6

    assert String not in pp_typeset.types

    assert String not in pp_typeset.relation_graph.nodes



from abc import ABCMeta, abstractmethod
from typing import Callable, Optional, Sequence, Type

import attr
import pandas as pd

from visions.relations import TypeRelation


class VisionsBaseTypeMeta(ABCMeta):
    def __contains__(cls, series: pd.Series, state: dict = {}) -> bool:
        # Possible alternative:
        # return cls in cls.typeset.detect_type_path(series)

        if series.empty:
            from visions.types import Generic

            return issubclass(cls, Generic)
        return cls.contains_op(series, state)  # type: ignore

    @property
    def relations(cls) -> Optional[Sequence[TypeRelation]]:
        if cls._relations is None:  # type: ignore
            cls._relations = cls.get_relations()  # type: ignore
        return cls._relations

    def __add__(cls, other):
        from visions.types import Generic
        from visions.typesets import VisionsTypeset

        if not any(issubclass(x, Generic) for x in [cls, other]):
            return VisionsTypeset([Generic, cls, other])
        return VisionsTypeset([cls, other])

    def __str__(cls) -> str:
        return str(cls.__name__)

    def __repr__(cls) -> str:
        return str(cls)


class VisionsBaseType(metaclass=VisionsBaseTypeMeta):
    """Abstract implementation of a vision type.

    Provides a common API for building custom visions data types.
    """

    _relations: Optional[Sequence[TypeRelation]] = None

    def __init__(self):
        raise ValueError("Types cannot be initialized")

    @classmethod
    @abstractmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        raise NotImplementedError

    @classmethod
    def evolve_type(
        cls,
        type_name: str,
        relations_generator: Optional[
            Callable[[Type[VisionsBaseTypeMeta]], Sequence[TypeRelation]]
        ] = None,
        replace: bool = False,
    ):
        """Make a copy of the type

        Args:
            type_name: the new type suffix, the type name will be `type[type_name]`
            relations_generator: a function returning all TypeRelations for the new type
            replace: if True, do not include the existing relations

        Returns:
            A new type
        """

        def get_new_relations(cls) -> Sequence[TypeRelation]:
            return relations

        new_type = type(
            "{name}[{type_name}]".format(name=cls.__name__, type_name=type_name),
            (cls,),
            {
                "get_relations": classmethod(get_new_relations),
                "contains_op": cls.contains_op,
            },
        )
        new_relations = (
            list(relations_generator(new_type)) if relations_generator else []
        )
        if replace:
            assert (
                relations_generator is not None
            ), "When calling evolve_type with `replace=True`, a `relations_generator` is required."
            relations = new_relations
        else:
            old_relations = [
                attr.evolve(relation, type=new_type) for relation in cls.get_relations()
            ]
            relations = old_relations + new_relations

        return new_type



from typing import List, Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils import func_nullable_series_contains


def to_int(series: pd.Series, state: dict) -> pd.Series:
    dtype = "Int64" if series.hasnans else np.int64
    return series.astype(dtype)


@func_nullable_series_contains
def float_is_int(series: pd.Series, state: dict) -> bool:
    def check_equality(series):
        try:
            if not np.isfinite(series).all():
                return False
            return series.eq(series.astype(int)).all()
        except:
            return False

    return check_equality(series)


def _get_relations(cls) -> List[TypeRelation]:
    from visions.types import Float, Generic

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(cls, Float, relationship=float_is_int, transformer=to_int),
    ]
    return relations


class Integer(VisionsBaseType):
    """**Integer** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1, 2, 3])
        >>> x in visions.Integer
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_integer_dtype(series)



from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Integer

    relations = [IdentityRelation(cls, Integer)]
    return relations


class Count(VisionsBaseType):
    """**Count** (positive integer) implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1, 4, 10, 20])
        >>> x in visions.Count
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_unsigned_integer_dtype(series)



import numpy as np
import pandas as pd
import pytest

from visions import StandardSet
from visions.test.series import get_geometry_series, get_series
from visions.test.utils import (
    cast,
    contains,
    convert,
    get_cast_cases,
    get_contains_cases,
    get_convert_cases,
    get_inference_cases,
    infers,
)
from visions.types import (
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Generic,
    Integer,
    Object,
    String,
    TimeDelta,
)
from visions.types.boolean import hasnan_bool_name

series = get_series() + get_geometry_series()

typeset = StandardSet()

contains_map = {
    Integer: {
        "int_series",
        "Int64_int_series",
        "int_range",
        "Int64_int_nan_series",
        "int_series_boolean",
        "np_uint32",
        "pd_uint32",
    },
    Float: {
        "float_series",
        "float_series2",
        "float_series3",
        "float_series4",
        "inf_series",
        "float_nan_series",
        "float_series5",
        "int_nan_series",
        "float_with_inf",
        "float_series6",
        # only NaNs
        "nan_series",
        "nan_series_2",
    },
    Categorical: {
        "categorical_int_series",
        "categorical_float_series",
        "categorical_string_series",
        "categorical_complex_series",
        "categorical_char",
        "ordinal",
    },
    Boolean: {"bool_series", "bool_series2", "bool_series3", "nullable_bool_series"},
    Complex: {
        "complex_series",
        "complex_series_py",
        "complex_series_nan",
        "complex_series_py_nan",
        "complex_series_nan_2",
        "complex_series_float",
    },
    DateTime: {
        "timestamp_series",
        "timestamp_aware_series",
        "datetime",
        "timestamp_series_nat",
        "date_series_nat",
    },
    TimeDelta: {"timedelta_series", "timedelta_series_nat", "timedelta_negative"},
    String: {
        "timestamp_string_series",
        "string_with_sep_num_nan",
        "string_series",
        "geometry_string_series",
        "string_unicode_series",
        "string_np_unicode_series",
        "path_series_linux_str",
        "path_series_windows_str",
        "int_str_range",
        "string_date",
        "textual_float",
        "textual_float_nan",
        "ip_str",
        "string_flt",
        "string_num",
        "str_url",
        "string_str_nan",
        "string_num_nan",
        "string_bool_nan",
        "string_flt_nan",
        "str_complex",
        "str_complex_nan",
        "uuid_series_str",
        "str_int_leading_zeros",
        "email_address_str",
        "str_float_non_leading_zeros",
        "str_int_zeros",
    },
}

if int(pd.__version__[0]) >= 1:
    contains_map[String].add("string_dtype_series")

contains_map[Object] = {
    "path_series_linux",
    "path_series_linux_missing",
    "path_series_windows",
    "url_series",
    "url_nan_series",
    "url_none_series",
    "file_test_py",
    "file_mixed_ext",
    "file_test_py_missing",
    "image_png",
    "image_png_missing",
    "image_png",
    "image_png_missing",
    "email_address",
    "email_address_missing",
    "uuid_series",
    "uuid_series_missing",
    "ip",
    "ip_mixed_v4andv6",
    "ip_missing",
    "geometry_series",
    "geometry_series_missing",
    "mixed_list[str,int]",
    "mixed_dict",
    "callable",
    "module",
    "mixed_integer",
    "mixed_list",
    "mixed",
    "bool_nan_series",
    "date",
    "time",
}

# Empty series
contains_map[Generic] = {
    "empty",
    "empty_bool",
    "empty_float",
    "empty_int64",
    "empty_object",
}


@pytest.mark.parametrize(**get_contains_cases(series, contains_map, typeset))
def test_contains(series, type, member):
    """Test the generated combinations for "series in type"

    Args:
        series: the series to test
        type: the type to test against
        member: the result
    """
    result, message = contains(series, type, member)
    assert result, message


inference_map = {
    "int_series": Integer,
    "categorical_int_series": Categorical,
    "int_nan_series": Integer,
    "Int64_int_series": Integer,
    "Int64_int_nan_series": Integer,
    "np_uint32": Integer,
    "pd_uint32": Integer,
    "int_range": Integer,
    "float_series": Float,
    "float_nan_series": Float,
    "int_series_boolean": Integer,
    "float_series2": Integer,
    "float_series3": Float,
    "float_series4": Float,
    "float_series5": Float,
    "float_series6": Float,
    "complex_series_float": Integer,
    "categorical_float_series": Categorical,
    "float_with_inf": Float,
    "inf_series": Float,
    "nan_series": Float,
    "nan_series_2": Float,
    "string_series": String,
    "categorical_string_series": Categorical,
    "timestamp_string_series": DateTime,
    "string_with_sep_num_nan": String,  # TODO: Introduce thousands separator
    "string_unicode_series": String,
    "string_np_unicode_series": String,
    "string_num_nan": Integer,
    "string_num": Integer,
    "string_flt_nan": Float,
    "string_flt": Float,
    "string_str_nan": String,
    "string_bool_nan": Boolean,
    "int_str_range": Integer,
    "string_date": DateTime,
    "str_url": String,
    "bool_series": Boolean,
    "bool_nan_series": Boolean,
    "nullable_bool_series": Boolean,
    "bool_series2": Boolean,
    "bool_series3": Boolean,
    "complex_series": Complex,
    "complex_series_nan": Complex,
    "complex_series_nan_2": Complex,
    "complex_series_py_nan": Complex,
    "complex_series_py": Complex,
    "categorical_complex_series": Categorical,
    "timestamp_series": DateTime,
    "timestamp_series_nat": DateTime,
    "timestamp_aware_series": DateTime,
    "datetime": DateTime,
    "timedelta_series": TimeDelta,
    "timedelta_series_nat": TimeDelta,
    "timedelta_negative": TimeDelta,
    "geometry_string_series": String,
    "geometry_series_missing": Object,
    "geometry_series": Object,
    "path_series_linux": Object,
    "path_series_linux_missing": Object,
    "path_series_linux_str": String,
    "path_series_windows": Object,
    "path_series_windows_str": String,
    "url_series": Object,
    "url_nan_series": Object,
    "url_none_series": Object,
    "mixed_list[str,int]": Object,
    "mixed_dict": Object,
    "mixed_integer": Object,
    "mixed_list": Object,
    "mixed": Boolean,
    "callable": Object,
    "module": Object,
    "textual_float": Float,
    "textual_float_nan": Float,
    "empty": Generic,
    "empty_object": Generic,
    "empty_float": Generic,
    "empty_bool": Generic,
    "empty_int64": Generic,
    "ip": Object,
    "ip_str": String,
    "ip_missing": Object,
    "date_series_nat": DateTime,
    "date": Object,
    "time": Object,
    "categorical_char": Categorical,
    "ordinal": Categorical,
    "str_complex": Complex,
    "str_complex_nan": Complex,
    "uuid_series": Object,
    "uuid_series_str": String,
    "uuid_series_missing": Object,
    "ip_mixed_v4andv6": Object,
    "file_test_py": Object,
    "file_test_py_missing": Object,
    "file_mixed_ext": Object,
    "image_png": Object,
    "image_png_missing": Object,
    "str_int_leading_zeros": String,
    "str_float_non_leading_zeros": Float,
    "str_int_zeros": Integer,
    "email_address": Object,
    "email_address_missing": Object,
    "email_address_str": String,
}
if int(pd.__version__[0]) >= 1:
    inference_map["string_dtype_series"] = String


@pytest.mark.parametrize(**get_inference_cases(series, inference_map, typeset))
def test_inference(series, type, typeset, difference):
    """Test the generated combinations for "inference(series) == type"

    Args:
        series: the series to test
        type: the type to test against
    """
    result, message = infers(series, type, typeset, difference)
    assert result, message


# Conversions in one single step
convert_map = [
    # Model type, Relation type
    (Integer, Float, {"int_nan_series", "float_series2"}),
    (Complex, String, {"str_complex", "str_complex_nan"}),
    (
        Float,
        String,
        {
            "string_flt",
            "string_num_nan",
            "string_num",
            "string_flt_nan",
            "textual_float",
            "textual_float_nan",
            "int_str_range",
            "str_float_non_leading_zeros",
            "str_int_zeros",
            # "string_with_sep_num_nan",
        },
    ),
    (DateTime, String, {"timestamp_string_series", "string_date"}),
    (Boolean, String, {"string_bool_nan"}),
    (Float, Complex, {"complex_series_float"}),
    (Boolean, Object, {"bool_nan_series", "mixed"}),
]


@pytest.mark.parametrize(**get_convert_cases(series, convert_map, typeset))
def test_conversion(source_type, relation_type, series, member):
    """Test the generated combinations for "convert(series) == type" and "infer(series) = source_type"

    Args:
        series: the series to test
        type: the type to test against
    """
    result, message = convert(source_type, relation_type, series, member)
    assert result, message


cast_results = {
    "float_series2": pd.Series([1, 2, 3, 4], dtype=np.int64),
    "int_nan_series": pd.Series([1, 2, np.nan], dtype=pd.Int64Dtype()),
    "timestamp_string_series": pd.Series(
        [np.datetime64("1941-05-24"), np.datetime64("2016-10-13")],
        dtype="datetime64[ns]",
    ),
    "string_num_nan": pd.Series([1, 2, np.nan], dtype=pd.Int64Dtype()),
    "string_num": pd.Series([1, 2, 3], dtype=np.int64),
    "string_flt_nan": pd.Series([1.0, 45.67, np.nan], dtype=np.float64),
    "string_flt": pd.Series([1.0, 45.67, 3.5], dtype=np.float64),
    "string_bool_nan": pd.Series([True, False, None], dtype=hasnan_bool_name),
    "int_str_range": pd.Series(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
        dtype=np.int64,
    ),
    "string_date": pd.Series(
        [np.datetime64("1937-05-06"), np.datetime64("2014-04-20")],
        dtype="datetime64[ns]",
    ),
    "str_float_non_leading_zeros": pd.Series([0.0, 0.04, 0.0], dtype=np.float64),
    "str_int_zeros": pd.Series([0, 0, 0, 2], dtype=np.int64),
    "bool_nan_series": pd.Series([True, False, None], dtype=hasnan_bool_name),
    "str_complex": pd.Series(
        [complex(1, 1), complex(2, 2), complex(10, 100)], dtype=np.complex128
    ),
    "str_complex_nan": pd.Series(
        [complex(1, 1), complex(2, 2), complex(10, 100), np.nan], dtype=np.complex128
    ),
    "complex_series_float": pd.Series([0, 1, 3, -1], dtype=np.int64),
    "textual_float": pd.Series([1.1, 2.0], dtype=np.float64),
    "textual_float_nan": pd.Series([1.1, 2.0, np.nan], dtype=np.float64),
    "mixed": pd.Series([True, False, None], dtype=hasnan_bool_name),
}


@pytest.mark.parametrize(**get_cast_cases(series, cast_results))
def test_cast(series, expected):
    if isinstance(expected, str):
        expected = None
    result, message = cast(series, typeset, expected)
    assert result, message



from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def to_ordinal(series: pd.Series) -> pd.Categorical:
    return pd.Categorical(
        series, categories=sorted(series.dropna().unique()), ordered=True
    )


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Categorical

    relations = [IdentityRelation(cls, Categorical)]
    return relations


class Ordinal(VisionsBaseType):
    """**Ordinal** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1, 2, 3, 1, 1], dtype='category')
        >>> x in visions.Ordinal
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_categorical_dtype(series) and series.cat.ordered



from datetime import date, time
from typing import Optional, Sequence

import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import (
    class_name_attrs,
    func_nullable_series_contains,
    nullable_series_contains,
)


@func_nullable_series_contains
def test_time(series: pd.Series) -> Optional[bool]:
    dtseries = series.dt.date
    value = date(1, 1, 1)
    return True if all(v == value for v in dtseries) else None


def to_time(series: pd.Series, state: dict) -> pd.Series:
    return series.dt.time


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object

    relations = [IdentityRelation(cls, Object)]
    return relations


class Time(VisionsBaseType):
    """**Time** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([datetime.time(10, 8, 4), datetime.time(21, 17, 0)])
        >>> x in visions.Time
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return class_name_attrs(series, time, ["microsecond", "hour"])



from visions.application.summaries import functional, summary
from visions.application.summaries.complete_summary import CompleteSummary
from visions.application.summaries.frame import *
from visions.application.summaries.series import *



import pandas as pd

from visions.relations.relations import InferenceRelation
from visions.relations.relations_utils import values_are_consecutive
from visions.types.ordinal import to_ordinal


def is_ordinal_str(s: pd.Series, state: dict) -> bool:
    if s.str.len().max() == 1:
        unique_values = list(s[s.notna()].str.lower().unique())
        return "a" in unique_values and values_are_consecutive(
            list(map(ord, unique_values))
        )
    else:
        return False


def string_to_ordinal(cls) -> InferenceRelation:
    from visions.types import String

    return InferenceRelation(
        type=cls,
        related_type=String,
        relationship=is_ordinal_str,
        transformer=to_ordinal,
    )



import numpy as np
import pandas as pd

from visions.relations.relations import InferenceRelation
from visions.types import Integer


def is_unsigned_int(series: pd.Series, state: dict) -> bool:
    # TODO: add coercion, ensure that > uint.MAX raises error
    return series.ge(0).all()


def to_unsigned_int(series: pd.Series) -> pd.Series:
    return series.astype(np.uint64)


def integer_to_count(cls) -> InferenceRelation:
    return InferenceRelation(
        relationship=is_unsigned_int,
        transformer=to_unsigned_int,
        type=cls,
        related_type=Integer,
    )



from pathlib import Path

from setuptools import find_packages, setup

# Read the contents of README file
source_root = Path(".")
with (source_root / "README.rst").open(encoding="utf-8") as f:
    long_description = f.read()

# Read the requirements
with (source_root / "requirements.txt").open(encoding="utf8") as f:
    requirements = f.readlines()

with (source_root / "requirements_dev.txt").open(encoding="utf8") as f:
    dev_requirements = f.readlines()

with (source_root / "requirements_test.txt").open(encoding="utf8") as f:
    test_requirements = f.readlines()

type_geometry_requires = ["shapely"]
type_image_path_requires = ["imagehash", "Pillow"]

extras_requires = {
    "type_geometry": type_geometry_requires,
    "type_image_path": type_image_path_requires,
    "plotting": ["pydot", "pygraphviz", "matplotlib"],
    "dev": dev_requirements,
    "test": test_requirements,
}

extras_requires["all"] = requirements + [
    dependency
    for name, dependencies in extras_requires.items()
    if name.startswith("type_") or name == "plotting"
    for dependency in dependencies
]


setup(
    name="visions",
    version="0.5.1",
    url="https://github.com/dylan-profiler/visions",
    description="Visions",
    author="Dylan Profiler",
    author_email="visions@ictopzee.nl",
    package_data={"vision": ["py.typed"]},
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=requirements,
    include_package_data=True,
    extras_require=extras_requires,
    tests_require=test_requirements,
    python_requires=">=3.6",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import Sequence

import attr
import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils
from visions.utils.series_utils import isinstance_attrs, nullable_series_contains


def string_is_email(series, state: dict):
    def test_email(s):
        return s.apply(str_to_email).apply(lambda x: x.local and x.fqdn)

    return test_utils.coercion_true_test(test_email)(series)


def str_to_email(s):
    if isinstance(s, FQDA):
        return s
    elif isinstance(s, str):
        return FQDA(*s.split("@", maxsplit=1))
    else:
        raise TypeError("Only strings supported")


def to_email(series: pd.Series, state: dict) -> pd.Series:
    return series.apply(str_to_email)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(
            cls,
            String,
            relationship=string_is_email,
            transformer=to_email,
        ),
    ]
    return relations


@attr.s(slots=True)
class FQDA(object):
    local = attr.ib()
    fqdn = attr.ib()

    @staticmethod
    def from_str(s):
        return str_to_email(s)


class EmailAddress(VisionsBaseType):
    """**EmailAddress** implementation of :class:`visions.types.type.VisionsBaseType`.

    Notes:
        The email address should be a **fully qualified domain address** (FQDA)
        FQDA = local part + @ + fully qualified domain name (FQDN)
        This type

    Examples:
        >>> x = pd.Series([FQDA('example','gmail.com'), FQDA.from_str('example@protonmail.com')])
        >>> x in visions.EmailAddress
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return isinstance_attrs(series, FQDA, ["local", "fqdn"])



import pandas as pd


def url_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {
        "scheme_counts": series.map(lambda x: x.scheme).value_counts(),
        "netloc_counts": series.map(lambda x: x.netloc).value_counts(),
        "path_counts": series.map(lambda x: x.path).value_counts(),
        "query_counts": series.map(lambda x: x.query).value_counts(),
        "fragment_counts": series.map(lambda x: x.fragment).value_counts(),
    }

    return summary



import pandas as pd


def range_summary(series: pd.Series) -> dict:
    """Summarize min, max and calculate the range

    Args:
        series: series to summarize

    Returns:
        A dict with `min`, `max` and `range`.
    """
    aggregates = ["min", "max"]
    summary = series.agg(aggregates).to_dict()

    summary["range"] = summary["max"] - summary["min"]
    return summary



from typing import Any, Tuple, Union

import numpy as np
import pandas as pd


def named_aggregate_summary(series: pd.Series, key: str):
    summary = {
        f"max_{key}": np.max(series),
        f"mean_{key}": np.mean(series),
        f"median_{key}": np.median(series),
        f"min_{key}": np.min(series),
    }

    return summary


def mad(arr, m=None):
    """Median Absolute Deviation: a "Robust" version of standard deviation.
    Indices variability of the sample.
    https://en.wikipedia.org/wiki/Median_absolute_deviation
    """
    if m is None:
        m = np.median(arr)
    return np.median(np.abs(arr - m))


def numerical_summary(
    series: pd.Series,
    quantiles=(0.05, 0.25, 0.5, 0.75, 0.95),
    count=None,
    is_unique=None,
    return_values=False,
) -> Union[dict, Tuple[dict, Any]]:
    """

    Args:
        series: series to summarize

    Returns:

    """

    if count is None:
        count = series.count()

    values = series.values
    present_values = values[~np.isnan(values)]
    finite_mask = np.isfinite(present_values)
    finite_values = present_values[finite_mask]

    summary = {
        "mean": np.mean(present_values),
        "std": np.std(present_values, ddof=1),
        "min": np.min(present_values),
        "max": np.max(present_values),
        # Unbiased kurtosis obtained using Fisher's definition (kurtosis of normal == 0.0). Normalized by N-1.
        "kurt": series.kurt(),
        # Unbiased skew normalized by N-1
        "skew": series.skew(),
        "sum": np.sum(present_values),
        "n_infinite": (~finite_mask).sum(),
        "n_zeros": (count - np.count_nonzero(present_values)),
    }

    for percentile, value in series.quantile(quantiles).to_dict().items():
        summary["quantile_{:d}".format(int(percentile * 100))] = value
    summary["median"] = summary["quantile_50"]
    summary["iqr"] = summary["quantile_75"] - summary["quantile_25"]

    summary["mad"] = mad(present_values, summary["quantile_50"])
    summary["variance"] = summary["std"] ** 2
    summary["cv"] = summary["std"] / summary["mean"] if summary["mean"] else np.NaN
    summary["range"] = summary["max"] - summary["min"]

    summary["monotonic_increase"] = series.is_monotonic_increasing
    summary["monotonic_decrease"] = series.is_monotonic_decreasing

    summary["monotonic_increase_strict"] = (
        summary["monotonic_increase"] and series.is_unique
    )
    summary["monotonic_decrease_strict"] = (
        summary["monotonic_decrease"] and series.is_unique
    )

    if return_values:
        return summary, finite_values

    return summary



from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic

    relations = [IdentityRelation(cls, Generic)]
    return relations


pandas_has_string_dtype_flag = hasattr(pdt, "is_string_dtype")


class Object(VisionsBaseType):
    """**Object** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series(['a', 1, np.nan])
        >>> x in visions.Object
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        is_object = pdt.is_object_dtype(series)
        if is_object:
            ret = True
        elif pandas_has_string_dtype_flag:
            ret = pdt.is_string_dtype(series) and not pdt.is_categorical_dtype(series)
        else:
            ret = False
        return ret



from visions.relations import IdentityRelation, TypeRelation
from visions.relations.integer_to_datetime import integer_to_datetime_year_month_day
from visions.types.date_time import DateTime
from visions.types.integer import Integer


def relations_generator(vtype):
    return [integer_to_datetime_year_month_day(vtype)]


def test_evolve_extend_datetime():
    DateTimeIntYYYYMMDD = DateTime.evolve_type("int_yyyymmdd", relations_generator)
    assert str(DateTimeIntYYYYMMDD) == "DateTime[int_yyyymmdd]"
    assert str(DateTime) == "DateTime"
    assert len(DateTime.get_relations()) == 2
    assert len(DateTimeIntYYYYMMDD.get_relations()) == 3


def test_evolve_replace_datetime():
    DateTimeIntYYYYMMDD = DateTime.evolve_type(
        "int_yyyymmdd_r", relations_generator, replace=True
    )
    assert str(DateTimeIntYYYYMMDD) == "DateTime[int_yyyymmdd_r]"
    assert str(DateTime) == "DateTime"
    assert len(DateTime.get_relations()) == 2
    assert len(DateTimeIntYYYYMMDD.get_relations()) == 1



import numpy as np
import pandas as pd

from visions import types as v
from visions.application.summaries import CompleteSummary

datetime_series = pd.Series(
    [
        pd.datetime(2010, 1, 1),
        pd.datetime(2010, 8, 2),
        pd.datetime(2011, 2, 1),
        np.datetime64("NaT"),
    ]
)

summarizer = CompleteSummary()
summary = summarizer.summarize_series(datetime_series, v.DateTime)
print(summary)

# Output:
# {
#     "dtype": dtype("<M8[ns]"),
#     "frequencies": {
#         Timestamp("2010-01-01 00:00:00"): 1,
#         Timestamp("2010-08-02 00:00:00"): 1,
#         Timestamp("2011-02-01 00:00:00"): 1,
#     },
#     "max": Timestamp("2011-02-01 00:00:00"),
#     "memory_size": 160,
#     "min": Timestamp("2010-01-01 00:00:00"),
#     "n_records": 4,
#     "n_unique": 3,
#     "na_count": 1,
#     "range": Timedelta("396 days 00:00:00"),
#     "types": {"NaTType": 1, "Timestamp": 3},
# }



from typing import Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils

# from visions.utils.series_utils import func_nullable_series_contains
from visions.utils.warning_handling import suppress_warnings


def test_string_leading_zeros(series: pd.Series, coerced_series: pd.Series):
    if coerced_series.hasnans:
        notna = coerced_series.notna()
        coerced_series = coerced_series[notna]

        if coerced_series.empty:
            return False
        series = series[notna]
    return not any(s[0] == "0" for s in series[coerced_series > 1])


# @func_nullable_series_contains
def string_to_float(series: pd.Series, state: dict) -> pd.Series:
    # Slightly faster to check for the character if it's not present than to
    # attempt the replacement
    # if any("," in x for x in series):
    #     series = series.str.replace(",", "")
    return series.astype(float)


def f(s):
    return s.astype(float)


def string_is_float(series: pd.Series, state: dict) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(f)(series)

    return (
        coerced_series is not None
        and coerced_series in Float
        and test_string_leading_zeros(series, coerced_series)
    )


def to_float(series: pd.Series, state: dict) -> pd.Series:
    return series.astype(float)


def complex_is_float(series, state: dict):
    return all(np.imag(series.values) == 0)


def complex_to_float(series, state: dict):
    return suppress_warnings(to_float)(series, state)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Complex, Generic, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls, String, relationship=string_is_float, transformer=string_to_float
        ),
        InferenceRelation(
            cls,
            Complex,
            relationship=complex_is_float,
            transformer=complex_to_float,
        ),
    ]
    return relations


class Float(VisionsBaseType):
    """**Float** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1.0, 2.5, 5.0, np.nan])
        >>> x in visions.Float
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_float_dtype(series)
